from typing import Final

from harbory.constants import DEFAULT_MLFACTORY_DIRECTORY, DEFAULT_WORKING_DIRECTORY

WORKFLOW_REFKEY: Final = "ref"
WORKFLOW_REFTYPE: Final = "ref"
WORKFLOW_DEFAULT_DIRECTORY: Final = DEFAULT_MLFACTORY_DIRECTORY / "workflow"
WORKFLOW_DEFAULT_SETTINGS_PATH: Final = DEFAULT_WORKING_DIRECTORY / "harbory.yml"
WORKFLOW_INTEGRATION_DIRECTORY: Final = DEFAULT_MLFACTORY_DIRECTORY / "integrations"
