from typing import Final

from harbory.workflow.constants import WORKFLOW_INTEGRATION_DIRECTORY

DEFAULT_MLFLOW_EXPERIMENT_NAME: Final = "Default"
DEFAULT_MLFLOW_DIRECTORY: Final = WORKFLOW_INTEGRATION_DIRECTORY / "mlflow"
